/***************************************************************************
 *   Copyright (C) 2008 by C J Plooy                                       *
 *   cornware-cjp@lycos.nl                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "kost_linalg.h"
#include "kost_constants.h"
#include "kost_math.h"

#include "kost_elements.h"

void kostElements2StateVector(kostReal mu, const kostElements *elements, kostStateVector *state)
{
	; /*TODO*/
}


/*Used to avoid division by zero etc.:*/
#define VERYSMALL 1e-6

void kostStateVector2Elements(kostReal mu, const kostStateVector *state, kostElements *elements, kostOrbitParam *params)
{
	/*
	See appendix C in orbiter.pdf
	*/

	kostVector3 h, n, e;
	kostReal absh, absn, absr, abse, E, tPe;
	int isEquatorial, isCircular, isHyperbola;

	if(params == NULL)
	{
		static kostOrbitParam dummy;
		params = &dummy;
	}

	h = kostCrossProductvv(&state->pos, &state->vel);
	n = kostConstructv(-h.y, h.x, 0.0);

	absh = kostAbsv(&h);
	absn = kostAbsv(&n);

	absr  = kostAbsv(&state->pos);

	/*
	Alternative formula for e:
	e = (v x h) / mu - r / |r|
	*/
	e = kostCrossProductvv(&state->vel, &h);
	e = kostMulrv((absr/mu), &e);
	e = kostSubvv(&e, &state->pos);
	e = kostMulrv((1.0/absr), &e);

	abse = kostAbsv(&e);

	isEquatorial = absn < VERYSMALL;
	isCircular   = abse < VERYSMALL;
	isHyperbola  = abse >= 1.0;

	E = 0.5*kostAbs2v(&state->vel) - mu / absr;

	/*SMa*/
	elements->a = -mu / (2.0*E);

	/*Ecc*/
	elements->e = abse;

	/*
	dp = a * (1-e)
	da = a * (1+e)
	*/
	params->PeD = elements->a * (1.0 - elements->e);
	params->ApD = elements->a * (1.0 + elements->e);

	/*Inc*/
	elements->i = acos(h.z / absh);

	/*LAN*/
	if(isEquatorial)
	{
		elements->theta = 0.0;
	}
	else
	{
		elements->theta = acos(n.x / absn);
		if(n.y < 0.0) elements->theta = M_TWOPI - elements->theta;
	}

	/*AgP*/
	params->AgP = 0.0;
	if(isCircular)
	{
		params->AgP = 0.0;
	}
	else if(isEquatorial)
	{
		params->AgP = acos(e.x / abse);
		if(e.z < 0.0) params->AgP = M_TWOPI - params->AgP;
	}
	else
	{
		params->AgP = acos(kostDotProductvv(&n, &e) / (absn*abse));
		if(e.z < 0.0) params->AgP = M_TWOPI - params->AgP;
	}

	/*TrA*/
	if(isCircular)
	{
		if(isEquatorial)
		{
			params->TrA = acos(state->pos.x / absr);
			if(state->vel.x > 0) params->TrA = M_TWOPI - params->TrA;
		}
		else
		{
			params->TrA = acos(kostDotProductvv(&n, &state->pos) / (absn*absr));
			if(kostDotProductvv(&n, &state->vel) > 0.0) params->TrA = M_TWOPI - params->TrA;
		}
	}
	else
	{
		kostReal tmp = kostDotProductvv(&e, &state->pos) / (abse*absr);

		/*Avoid acos out of range:*/
		if(tmp <= -1.0)
			{params->TrA = M_PI;}
		else if(tmp >= 1.0)
			{params->TrA = 0.0;}
		else
			{params->TrA = acos(tmp);}

		if(kostDotProductvv(&state->pos, &state->vel) < 0.0) params->TrA = M_TWOPI - params->TrA;
	}

	/*Lec*/
	params->Lec = elements->a * elements->e;

	/*
	SMi
	b^2 = a^2 * (1 - e^2)
	*/
	if(isHyperbola)
	{
		params->SMi =
			sqrt(elements->a*elements->a* (elements->e*elements->e - 1.0));
	}
	else
	{
		params->SMi =
			sqrt(elements->a*elements->a* (1.0 - elements->e*elements->e));
	}

	/*LPe*/
	elements->omegab = elements->theta + params->AgP;
	while(elements->omegab > M_TWOPI) elements->omegab -= M_TWOPI;

	/*EcA*/
	if(isHyperbola)
	{
		params->EcA = acosh((1.0 - absr/elements->a) / elements->e);
	}
	else if(isCircular)
	{
		params->EcA = 0.0;
	}
	else
	{
		kostReal tmp = (1.0 - absr/elements->a) / elements->e;

		/*Avoid acos out of range:*/
		if(tmp <= -1.0)
			{params->EcA = M_PI;}
		else if(tmp >= 1.0)
			{params->EcA = 0.0;}
		else
			{params->EcA = acos(tmp);}
	}

	if(isHyperbola)
	{
		/*Copy sign from sin(TrA)*/
		if(sin(params->TrA) * params->EcA < 0.0)
			params->EcA = -params->EcA;
	}
	else
	{
		/*Same rule basically, but with EcA in 0..2pi range*/
		if(sin(params->TrA) < 0.0) params->EcA = M_TWOPI - params->EcA;
	}

	/*MnA*/
	if(isHyperbola)
	{
		params->MnA = elements->e * sinh(params->EcA) - params->EcA;
	}
	else
	{
		params->MnA = params->EcA - elements->e * sin(params->EcA);
	}

	/*MnL*/
	elements->L = params->MnA + elements->omegab;
	while(elements->L > M_TWOPI) elements->L -= M_TWOPI;

	/*TrL*/
	params->TrL = elements->omegab + params->TrA;
	while(params->TrL > M_TWOPI) params->TrL -= M_TWOPI;

	/*
	T = 2*pi*sqrt(a^3 / mu)

	fabs is for supporting hyperbola
	*/
	params->T = 
		M_TWOPI * sqrt(fabs(elements->a*elements->a*elements->a/mu));

	/*
	Calculating PeT and ApT:
	*/
	tPe = params->MnA * params->T / M_TWOPI; /*Time since last Pe*/

	if(isHyperbola)
	{
		params->PeT = -tPe;
	}
	else
	{
		params->PeT = params->T - tPe;
	}

	params->ApT = 0.5*params->T - tPe;
	if(params->ApT < 0.0) params->ApT += params->T;
}
