/***************************************************************************
 *   Copyright (C) 2008 by C J Plooy                                       *
 *   cornware-cjp@lycos.nl                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#define STRICT
#define ORBITER_MODULE

#include <cstdio>
#include <cstdlib>

#include <algorithm> //for std::swap
#include <vector>

#include <windows.h>

#include "orbitersdk.h"

#include "popupmenu.h"
#include "kost.h"

#include "planettree.h"
#include "colors.h"

#include "lagrangemfd.h"

bool rf1MenuResponse(void *menuPtr, unsigned int menuIndex, char *name, void *userPtr)
{
	CLagrangeMFD *mfd = (CLagrangeMFD *)userPtr;

	OBJHANDLE newRef = oapiGetGbodyByName(name);

	if(newRef == NULL) return false; //Reject non-existent objects

	//Reject non-celestial body objects:
	int objType = oapiGetObjectType(newRef);
	if(objType != OBJTP_STAR && objType != OBJTP_PLANET) return false;

	//Assign reference:
	mfd->m_Reference1 = newRef;
	mfd->updateReferenceOrbit();
	mfd->InvalidateDisplay();

	return true; //accept selection
}

bool rf2MenuResponse(void *menuPtr, unsigned int menuIndex, char *name, void *userPtr)
{
	CLagrangeMFD *mfd = (CLagrangeMFD *)userPtr;

	OBJHANDLE newRef2 = oapiGetGbodyByName(name);

	if(newRef2 == NULL) return false; //Reject non-existent objects

	//Reject non-celestial body objects:
	int objType = oapiGetObjectType(newRef2);
	if(objType != OBJTP_STAR && objType != OBJTP_PLANET) return false;

	//Assign reference:
	mfd->m_Reference2 = newRef2;
	mfd->updateReferenceOrbit();
	mfd->InvalidateDisplay();

	return true; //accept selection
}

bool tgtMenuResponse(void *menuPtr, unsigned int menuIndex, char *name, void *userPtr)
{
	CLagrangeMFD *mfd = (CLagrangeMFD *)userPtr;

	OBJHANDLE newTgt = oapiGetObjectByName(name);

	if(newTgt == NULL) return false; //Reject non-existent objects

	//Assign target:
	mfd->m_Target = newTgt;
	mfd->InvalidateDisplay();

	return true; //accept selection
}

int LagrangeMFDMode = 0;

/*
Static storage of MFD state:
*/
struct
{
	OBJHANDLE m_Reference1, m_Reference2;
	OBJHANDLE m_Target;

	double m_ReferenceSimTime;
	double m_ReferencePeriod;
	double m_ReferenceBarycentre;
	kostMatrix3 m_ReferencePlane;

} _LagrangeMFDState;

DLLCLBK void InitModule (HINSTANCE hModule)
{
	static char *name = "Lagrange MFD";
	MFDMODESPEC spec;
	spec.name    = name;
	spec.key     = OAPI_KEY_F; // No longer used according to Orbiter docs
	spec.msgproc = CLagrangeMFD::MsgProc;

	LagrangeMFDMode = oapiRegisterMFDMode(spec);

	//Initial "un-initialized" state:
	_LagrangeMFDState.m_Reference1 = NULL;
	_LagrangeMFDState.m_Reference2 = NULL;
	_LagrangeMFDState.m_Target    = NULL;

	//Load user-defined colors:
	loadColorConfiguration();
}

DLLCLBK void ExitModule (HINSTANCE hModule)
{
	oapiUnregisterMFDMode(LagrangeMFDMode);
}

CLagrangeMFD::CLagrangeMFD(DWORD w, DWORD h, VESSEL *vessel) : MFD (w, h, vessel)
{
	m_Width = w;
	m_Height = h;
	m_Vessel = vessel;

	m_LastTraceSimTime = 0.0;

	//Initial mode:
	m_Target     = NULL;

	//Initialize planet tree data (TODO: find better place for this)
	initPlanetTree();

	//Initialize reference bodies:
	autoReference();
}

CLagrangeMFD::~CLagrangeMFD()
{
}

int CLagrangeMFD::MsgProc(UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
	case OAPI_MSG_MFD_OPENED:
		return (int)(new CLagrangeMFD(LOWORD(wparam), HIWORD(wparam), (VESSEL*)lparam));
	}
	return 0;
}

void CLagrangeMFD::WriteStatus(FILEHANDLE scn) const
{
	//oapiWriteScenario_xxx

	char name[256];

	oapiGetObjectName(m_Reference1, name, 255);
	oapiWriteScenario_string(scn, "REF1", name);

	if(m_Reference2 != NULL)
	{
		oapiGetObjectName(m_Reference2, name, 255);
		oapiWriteScenario_string(scn, "REF2", name);
	}

	if(m_Target != NULL)
	{
		oapiGetObjectName(m_Target, name, 255);
		oapiWriteScenario_string(scn, "TARGET", name);
	}
}

void CLagrangeMFD::ReadStatus(FILEHANDLE scn)
{
	//oapiReadScenario_nextline

	//Default:
	m_Target     = NULL;

	char *line = NULL;
	while(oapiReadScenario_nextline(scn, line))
	{
		unsigned int len = strlen(line);

		if(len > 5 && strncmp(line, "REF1 ", 5) == 0)
		{
			char *refname = line+5;
			OBJHANDLE newRef = 	oapiGetGbodyByName(refname);

			if(newRef == NULL) continue; //Reject non-existent objects

			//Reject non-celestial body objects:
			int objType = oapiGetObjectType(newRef);
			if(objType != OBJTP_STAR && objType != OBJTP_PLANET) continue;

			//Assign reference:
			m_Reference1 = newRef;
		}

		if(len > 5 && strncmp(line, "REF2 ", 5) == 0)
		{
			char *refname = line+5;
			OBJHANDLE newRef = 	oapiGetGbodyByName(refname);

			if(newRef == NULL) continue; //Reject non-existent objects

			//Reject non-celestial body objects:
			int objType = oapiGetObjectType(newRef);
			if(objType != OBJTP_STAR && objType != OBJTP_PLANET) continue;

			//Assign reference:
			m_Reference2 = newRef;
		}

		if(len > 7 && strncmp(line, "TARGET ", 7) == 0)
		{
			char *tgtname = line+7;
			OBJHANDLE newTgt = oapiGetObjectByName(tgtname);

			if(newTgt == NULL) continue; //Reject non-existent objects

			//Assign target:
			m_Target = newTgt;
		}
	}

	if(m_Reference1 != NULL && m_Reference2 != NULL)
		updateReferenceOrbit();
}

void CLagrangeMFD::StoreStatus() const
{
	_LagrangeMFDState.m_Reference1 = m_Reference1;
	_LagrangeMFDState.m_Reference2 = m_Reference2;
	_LagrangeMFDState.m_Target     = m_Target;

	_LagrangeMFDState.m_ReferenceSimTime    = m_ReferenceSimTime;
	_LagrangeMFDState.m_ReferencePeriod     = m_ReferencePeriod;
	_LagrangeMFDState.m_ReferenceBarycentre = m_ReferenceBarycentre;
	_LagrangeMFDState.m_ReferencePlane      = m_ReferencePlane;
}

void CLagrangeMFD::RecallStatus()
{
	if(_LagrangeMFDState.m_Reference1 == NULL) return; //Static data not initialized

	m_Reference1 = _LagrangeMFDState.m_Reference1;
	m_Reference2 = _LagrangeMFDState.m_Reference2;
	m_Target     = _LagrangeMFDState.m_Target;

	m_ReferenceSimTime    = _LagrangeMFDState.m_ReferenceSimTime;
	m_ReferencePeriod     = _LagrangeMFDState.m_ReferencePeriod;
	m_ReferenceBarycentre = _LagrangeMFDState.m_ReferenceBarycentre;
	m_ReferencePlane      = _LagrangeMFDState.m_ReferencePlane;
}

// Return button labels
char *CLagrangeMFD::ButtonLabel(int bt)
{
	// The labels for the two buttons used by our MFD mode
	static char *label[7] =
	{"RF1", "RF2", "AR", "TGT", "NT", "UPD", "CLR"};

	return (bt < 7 ? label[bt] : NULL);
}

// Return button menus
int CLagrangeMFD::ButtonMenu(const MFDBUTTONMENU **menu) const
{
	// The menu descriptions for the two buttons
	static const MFDBUTTONMENU mnu[7] = {
		{"Orbit reference 1",NULL,              '1'},
		{"Orbit reference 2",NULL,              '2'},
		{"Auto reference",   NULL,              'A'},
		{"Select target",    NULL,              'T'},
		{"Unselect target",  NULL,              'N'},
		{"Update reference", "orbit",           'U'},
		{"Clear traces",     NULL,              'C'}
	};
	if (menu) *menu = mnu;

	return 7; // return the number of buttons used
}

bool CLagrangeMFD::ConsumeButton(int bt, int event)
{
	static DWORD keys[7] =
	{OAPI_KEY_1, OAPI_KEY_2, OAPI_KEY_A, OAPI_KEY_T, OAPI_KEY_N, OAPI_KEY_U, OAPI_KEY_C};

	if(bt >= 7) return false;
	if(!(event & PANEL_MOUSE_LBDOWN)) return false;

	return ConsumeKeyBuffered(keys[bt]);
}

bool CLagrangeMFD::ConsumeKeyBuffered(DWORD key)
{
	switch(key)
	{
	case OAPI_KEY_1:
		btnRF1();
		return true;
	case OAPI_KEY_2:
		btnRF2();
		return true;
	case OAPI_KEY_A:
		btnAR();
		return true;
	case OAPI_KEY_T:
		btnTGT();
		return true;
	case OAPI_KEY_N:
		btnNT();
		return true;
	case OAPI_KEY_U:
		btnUPD();
		return true;
	case OAPI_KEY_C:
		btnCLR();
		return true;
	}

	return false;
}

void CLagrangeMFD::btnRF1()
{
	oapiCelBodyPopup(
		this, "Lagrange MFD: Reference 1", rf1MenuResponse, 0);
}

void CLagrangeMFD::btnRF2()
{
	oapiTargetPopup(
		this, "Lagrange MFD: Reference 2", rf2MenuResponse, m_Reference1, 0);
}

void CLagrangeMFD::btnAR()
{
	autoReference();
	InvalidateDisplay();
}

void CLagrangeMFD::btnTGT()
{
	oapiTargetPopup(this, "Lagrange MFD: Target", tgtMenuResponse, m_Reference1, 0);
}

void CLagrangeMFD::btnNT()
{
	m_Target = NULL;
	InvalidateDisplay();
}

void CLagrangeMFD::btnUPD()
{
	updateReferenceOrbit();
	InvalidateDisplay();
}

void CLagrangeMFD::btnCLR()
{
	clearTraces();
	InvalidateDisplay();
}

void CLagrangeMFD::autoReference()
{
	VESSELSTATUS status;
	m_Vessel->GetStatus(status);
	m_Reference1 = status.rbody;

	m_Reference2 = NULL;

	//Find a suitable child body:
	double maxMass = 0.0;

	unsigned int numBodies = oapiGetGbodyCount();
	unsigned int i = 0;
	for(i=0; i < numBodies; i++)
	{
		OBJHANDLE body = oapiGetGbodyByIndex(i);

		if(body == m_Reference1) continue;

		//Parent test
		OBJHANDLE parent = getCentralBody(body); //From planet tree
		if(parent != m_Reference1) continue;

		//Max mass selection
		double m = oapiGetMass(body);
		if(m < maxMass) continue;
		maxMass = m;

		m_Reference2 = body;
	}

	if(m_Reference2 == NULL)
	{
		//Suitable child body not found, looking for parent body
		m_Reference2 = m_Reference1;

		m_Reference1 = getCentralBody(m_Reference2);

		if(m_Reference1 == NULL)
		{
			//No parent body found: fall back to safe un-initialized mode
			m_Reference1 = m_Reference2;
			m_Reference2 = NULL;
		}
	}

	updateReferenceOrbit();
}

void CLagrangeMFD::updateReferenceOrbit()
{
	double m1 = oapiGetMass(m_Reference1);
	double m2 = oapiGetMass(m_Reference2);

	m_ReferenceBarycentre = m2 / (m1+m2);

	kostReal mu = GRAVITATIONAL_CONSTANT * (m1+m2);

	kostStateVector state;
	oapiGetRelativePos(m_Reference2, m_Reference1, &state.pos);
	oapiGetRelativeVel(m_Reference2, m_Reference1, &state.vel);

	//Transform to standard right-handed coordinate system:
	//y -> z
	//z -> y
	std::swap<kostReal>(state.pos.y, state.pos.z);
	std::swap<kostReal>(state.vel.y, state.vel.z);

	//Calculate orbital elements & params
	kostElements elements;
	kostOrbitParam params;
	kostStateVector2Elements(mu, &state, &elements, &params);

	m_ReferencePeriod = params.T;
	m_ReferenceSimTime = oapiGetSimTime() + params.PeT;


	kostMatrix3 AgPMat, LANMat, IncMat;
	kostMakeZRotm(&AgPMat, params.AgP);
	kostMakeXRotm(&IncMat, elements.i);
	kostMakeZRotm(&LANMat, elements.theta);

	// Now, global = LANMat * IncMat * AgPMat * local:
	m_ReferencePlane = LANMat * IncMat * AgPMat;

	//From global to local:
	makeTranspose(m_ReferencePlane);
}

void CLagrangeMFD::updatePositions()
{
	//Get relative to ref1:
	oapiGetRelativePos(m_Reference2, m_Reference1, &m_Ref2pos);
	if(m_Target != NULL)
		oapiGetRelativePos(m_Target, m_Reference1, &m_Targetpos);
	m_Vessel->GetRelativePos(m_Reference1, m_Vesselpos);
	m_Vessel->GetRelativeVel(m_Reference1, m_Vesselvel);

	//Transform to standard right-handed coordinate system:
	//y -> z
	//z -> y
	std::swap<kostReal>(m_Ref2pos.y  , m_Ref2pos.z  );
	std::swap<kostReal>(m_Targetpos.y, m_Targetpos.z);
	std::swap<kostReal>(m_Vesselpos.y, m_Vesselpos.z);
	std::swap<kostReal>(m_Vesselvel.y, m_Vesselvel.z);

	//Make relative to barycentre:
	kostVector3 barycentre = m_Ref2pos * m_ReferenceBarycentre;
	kostVector3 baryvel;
	oapiGetRelativeVel(m_Reference2, m_Reference1, &baryvel);
	baryvel *= m_ReferenceBarycentre;

	m_Ref1pos    = barycentre * -1.0;
	m_Ref2pos   -= barycentre;
	m_Targetpos -= barycentre;
	m_Vesselpos -= barycentre;
	m_Vesselvel -= baryvel;

	//Rotating reference according to sim time
	double t = oapiGetSimTime() - m_ReferenceSimTime;
	kostReal angle = M_TWOPI * t / m_ReferencePeriod;
	kostMatrix3 transform;
	kostMakeZRotm(&transform, -angle);

	transform = transform * m_ReferencePlane;

	//Transform points:
	m_Ref1pos   *= transform;
	m_Ref2pos   *= transform;
	m_Targetpos *= transform;
	m_Vesselpos *= transform;
	m_Vesselvel *= transform;

	//Subtract rotational velocity from velocities
	kostVector3 omega = kostConstructv(0.0, 0.0, M_TWOPI / m_ReferencePeriod);
	m_Vesselvel -= crossProduct(omega, m_Vesselpos);
}

void CLagrangeMFD::clearTraces()
{
	m_Ref1Trace.clear();
	m_Ref2Trace.clear();
	m_VesselTrace.clear();
	m_TargetTrace.clear();
}

void CLagrangeMFD::updateTraces()
{
	const unsigned int maxTracePoints = 256;

	double dt = m_ReferencePeriod / maxTracePoints;
	double t = oapiGetSimTime();

	if(t - m_LastTraceSimTime > dt)
	{
		//Remove first trace point(s) if necessary
		while(m_Ref1Trace.size() >= maxTracePoints)
		{
			m_Ref1Trace.pop_front();
			m_Ref2Trace.pop_front();
			m_VesselTrace.pop_front();
			m_TargetTrace.pop_front();
		}

		//Add trace points
		m_Ref1Trace.push_back  (m_Ref1pos);
		m_Ref2Trace.push_back  (m_Ref2pos);
		m_VesselTrace.push_back(m_Vesselpos);
		m_TargetTrace.push_back(m_Targetpos);

		m_LastTraceSimTime = t;
	}
}

void CLagrangeMFD::updateLagrangePoints()
{
	double m1 = oapiGetMass(m_Reference1);
	double m2 = oapiGetMass(m_Reference2);
	kostVector3 r1 = m_Ref1pos;
	kostVector3 r2 = m_Ref2pos;

	kostVector3 dr = r2-r1;
	double R = abs(dr);

	kostVector3 xdir = normal(r2-r1);
	kostVector3 zdir = kostConstructv(0,0,1);
	kostVector3 ydir = crossProduct(zdir, xdir);

	//L4 and L5 points (exact positions):
	m_L[3] = r1 + xdir*(0.5*R) + ydir*(0.5*sqrt(3.0)*R);
	m_L[4] = r1 + xdir*(0.5*R) - ydir*(0.5*sqrt(3.0)*R);

	//L1..L3 (approximate positions):
	double hillSphereR = R*pow(m2/(3.0*m1),(1.0/3.0));
	m_L[0] = r2 - xdir*hillSphereR;
	m_L[1] = r2 + xdir*hillSphereR;
	m_L[2] = r1 - dr;

	//TODO: refine L1..L3
}

void CLagrangeMFD::Update(HDC hDC)
{
	//Title
	{
		char refname1[10], refname2[10];
		//Initialize as empty string:
		refname1[0] = '\0';
		refname2[0] = '\0';
		//Get names
		if(m_Reference1 != NULL) oapiGetObjectName(m_Reference1, refname1, 9);
		if(m_Reference2 != NULL) oapiGetObjectName(m_Reference2, refname2, 9);

		//Make sure it is NUL-terminated:
		refname1[9] = '\0';
		refname2[9] = '\0';

		//The title:
		char title[36];
		sprintf(title, "Lagrange: %s & %s", refname1, refname2);
		Title(hDC, title);
	}

	//Don't display until both are set:
	if(m_Reference1 == NULL || m_Reference2 == NULL) return;

	updatePositions();
	updateTraces();
	updateLagrangePoints();

	calcScaleAndRange();

	draw(hDC);
}

void CLagrangeMFD::calcScaleAndRange()
{
	//Some points that should be inside the view:
	std::vector<kostVector3> points;
	unsigned int i=0;

	//We're looking at 2D coordinates only, so
	//3D relationships like |PeD| <= |Dst| can not
	//be used.

	points.push_back(m_Ref1pos);
	points.push_back(m_Ref2pos);
	points.push_back(m_Vesselpos);
	if(m_Target != NULL)
		points.push_back(m_Targetpos);
	for(i=0; i < 5; i++)
		points.push_back(m_L[i]);

	double largest = 0.0;
	for(i=0; i < points.size(); i++)
	{
		kostVector3 &vec = points[i];
		double len_2d = vec.x*vec.x + vec.y*vec.y;
		if(len_2d > largest) largest = len_2d;
	}
	largest = sqrt(largest); //Turn length^2 into length

	m_Scale = 0.45 * double(m_Width) / largest;
}

void CLagrangeMFD::draw(HDC hDC)
{
	//This allows SetDCPenColor to work:
	SelectObject(hDC, GetStockObject(DC_PEN));

	//Reference bodies
	SetDCPenColor(hDC, OrbitMFDPlanetColor);
	drawTracePoint(hDC, m_Ref1Trace, m_Ref1pos);
	drawTracePoint(hDC, m_Ref2Trace, m_Ref2pos);

	//Lagrange points
	for(unsigned int i=0; i<5; i++)
		drawPoint(hDC, m_L[i], 1);

	//Vessel
	SetDCPenColor(hDC, OrbitMFDVesselOrbitColor);
	drawTracePoint(hDC, m_VesselTrace, m_Vesselpos);

	//Target
	if(m_Target != NULL)
	{
		SetDCPenColor(hDC, OrbitMFDTargetOrbitColor);
		drawTracePoint(hDC, m_TargetTrace, m_Targetpos);
	}

	//Velocity indicator:
	{
		SetTextColor(hDC, OrbitMFDVesselOrbitColor);

		char veltxt[80];

		unsigned int x = 3;
		unsigned int y = 16;
		unsigned int dy = m_Height/22;

		double totvel = abs(m_Vesselvel);
		sprintf(veltxt, "Tot vel: %f", totvel);
		TextOut(hDC, x, y, veltxt, strlen(veltxt));
		y += dy;

		kostVector3 posdir = normal(m_Vesselpos);
		double radvel = dotProduct(m_Vesselvel, posdir);
		sprintf(veltxt, "Rad vel: %f", radvel);
		TextOut(hDC, x, y, veltxt, strlen(veltxt));
		y += dy;

		kostVector3 tandir = crossProduct(kostConstructv(0,0,1), posdir);
		double tanvel = dotProduct(m_Vesselvel, tandir);
		sprintf(veltxt, "Tan vel: %f", tanvel);
		TextOut(hDC, x, y, veltxt, strlen(veltxt));
		y += dy;

		kostVector3 zdir = kostConstructv(0,0,1);
		double zvel = dotProduct(m_Vesselvel, zdir);
		sprintf(veltxt, "Z   vel: %f", zvel);
		TextOut(hDC, x, y, veltxt, strlen(veltxt));
		y += dy;
	}
}

void CLagrangeMFD::drawTracePoint(HDC hDC, std::list<kostVector3> &trace, const kostVector3 &vec) const
{
	trace.push_back(vec);
	drawTrace(hDC, trace);
	trace.pop_back();

	drawPoint(hDC, vec);
}

void CLagrangeMFD::drawPoint(HDC hDC, const kostVector3 &vec, unsigned int size) const
{
	POINT pt = real2screen(vec);
	Rectangle(hDC, pt.x-size, pt.y-size, pt.x+size, pt.y+size);
}

void CLagrangeMFD::drawTrace(HDC hDC, const std::list<kostVector3> &trace) const
{
	unsigned int numPts = trace.size();
	if(numPts < 2) return;

	POINT *points = new POINT[numPts];

	unsigned int n = 0;
	std::list<kostVector3>::const_iterator i;
	for(i=trace.begin(); i != trace.end(); i++)
	{
		points[n] = real2screen(*i);
		n++;
	}

	Polyline(hDC, points, numPts);
	delete [] points;
}

POINT CLagrangeMFD::real2screen(const kostVector3 &vec) const
{
	kostVector3 pt = vec * m_Scale;
	POINT ret;
	ret.x = m_Width/2 + int(pt.x);
	ret.y = m_Height/2 - int(pt.y); //Flipped y-axis
	return ret;
}

